__authors__ = "H. Andrew Schwartz, Salvatore Giorgi, Maarten Sap, Patrick Crutchley, Lukasz Dziurzynski and Megha Agrawal"
__copyright__ = "Copyright 2017"
__credits__ = []
__license__ = "Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License: http://creativecommons.org/licenses/by-nc-sa/3.0/"
__version__ = "1.1.4" 
__maintainer__ = "Salvatore Giorgi, Andy Schwartz"
__email__ = "sgiorgi@sas.upenn.edu, has@cs.stonybrook.edu"
